use v6;

module File::Find;

class File::Find::Result is Cool {
	has $.dir;
	has $.name;

	method Str {
		$.dir ~ '/' ~ $.name
	}
}

sub find (:$dir!, :$name) is export {
	my @targets = dir($dir).map: {
		File::Find::Result.new(dir => $dir, name => $_);
	};
	my $list = gather while @targets {
		my $elem = @targets.shift;
		if $name.defined {
			take $elem if $elem ~~ $name;
		} else {
			take $elem;
		}
		if $elem.IO ~~ :d {
			for dir($elem) -> $file {
				@targets.push(
					File::Find::Result.new(dir => $elem, name => $file)
				);
			}
		}
	}
	return $list;
}

=begin pod

=head1 NAME

File::Find - Get a lazy list of a directory tree

=head1 SYNOPSIS

	use File::Find;

	my @list := find(dir => 'foo');
	say @list[0..3];

	my $list = find(dir => 'foo');
	say $list[0..3];

=head1 DESCRIPTION

C<File::Find> allows you to get the contents of the given directory,
recursively. The only exported function, C<find()>, generates a lazy
list of files in given directory. Every element of the list is a
C<File::Find::Result> object, described below.
C<find()> takes the following (named) arguments:

=head2 dir (B<required>)

The directory which C<find()> will traverse

=head2 name

The regex the potential file will be matched against. If the match
succeeds, the file will be added to a returned list. You can also pass
a string here, but note that it will match the whole path against it,
not just the filename. To look for a certain filename, You'll probably
want something like C<<find(dir => 'foo' name => / \/file.name$ /)>>.

=head1 File::Find::Result

C<File::Find::Result> object acts like a normal string, having two
additional accessors, C<dir> and C<name>, holding the directory
the file is in and the filename respectively.

=head1 Perl 5's File::Find

Please note, that this module is not trying to be the verbatim port of
Perl 5's File::Find module. Its interface is closer to Perl 5's
File::Find::Rule, and its features are planned to be similar one day.

=head1 CAVEATS

List assignment is eager in Perl 6, so if You assign C<find()> result
to an array, the elements will be copied and the laziness will be
spoiled. For a proper lazy list, use either binding (C<:=>) or assign
a result to a scalar value (see SYNOPSIS).

=end pod

# vim: ft=perl6
